﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  /// <summary>
  /// An operator that analyzes the quality of schemas.
  /// </summary>
  [Item("BestIndividualWildcardAnalyzer", "An operator that analyzes unbroken upwards inheritance of schemata.")]
  [StorableClass]
  public sealed class BestIndividualWildcardAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultName = "PopulationGraph";
    private const string BestIndividualSchemataResultName = "BestIndividualWildcards";
    private const string UnbrokenUpwardsInteritanceWildcardOccurrenceMatrixResultName = "UnbrokenUpwardsInteritanceWildcardOccurrenceMatrix";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private BestIndividualWildcardAnalyzer(bool deserializing) : base(deserializing) { }
    private BestIndividualWildcardAnalyzer(BestIndividualWildcardAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public BestIndividualWildcardAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestIndividualWildcardAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var schemata = (ItemArray<IntArray>)Results[BestIndividualSchemataResultName].Value;
      if (schemata.Count() > 1) {
        var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultName].Value;
        int maxGen = (int)graph.Values.Max(x => x.Rank.Max()); // find last generation
        SchemaEqualityComparer comparer = new SchemaEqualityComparer();

        double curMaxGen = maxGen;
        IEnumerable<GenealogyGraphNode> lastGen = null;
        IEnumerable<GenealogyGraphNode> parents = null;
        lastGen = graph.Values.Where(x => x.Rank.Contains(curMaxGen));
        parents = lastGen.Where(x => comparer.Equals((IntArray)x.Data, schemata.Last()));

        while (parents.Count() == 0) {
          lastGen = graph.Values.Where(x => x.Rank.Contains(curMaxGen));
          parents = lastGen.Where(x => comparer.Equals((IntArray)x.Data, schemata.Last()));
          curMaxGen -= 0.5;
        }
        var bestIndividual = parents.First();
        var ancestors = parents.First().Ancestors().Where(x => AlgorithmBehaviorHelpers.IsSubtour(schemata.Last(), (Permutation)x.Data));



        // locate best individual in graph
        /*GenealogyGraphNode bestIndividual = null;
        try {
          var parents = lastGen.Where(x => comparer.Equals((IntArray)x.Data, parent));
          bestIndividual = graph.Values.First(x => x.Rank.Max() == maxGen && AlgorithmBehaviorHelpers.IsMatch(new IntArray(((Permutation)x.Data).ToArray()), (schemata.Last()).ToArray()));
        }
        catch (Exception ex) {
          ex.ToString();
        } */
        var bestSchemata = schemata.Take(schemata.Count() - 1);

        // group ancestors by generation
        var generations = new IList<GenealogyGraphNode>[((int)maxGen) + 1];
        for (int i = 0; i <= maxGen; i++)
          generations[i] = bestIndividual.Ancestors().Where(x => x.Rank.Contains(i) || x.Rank.Contains(i - 0.5)).ToList();


        var unbrokenUpwardsInteritanceSchemaOccurrenceMatrix = new StringMatrix(bestSchemata.Count(), 1);
        IList<string> rowNames = new List<string>();
        for (int i = 0; i < bestSchemata.Count(); i++) {
          var schema = bestSchemata.ElementAt(i);
          rowNames.Add(new Permutation(PermutationTypes.RelativeUndirected, schema).ToString());
          unbrokenUpwardsInteritanceSchemaOccurrenceMatrix[i, 0] = TrackUnbrokenAncestorSchemaPath(schema, bestIndividual, generations);
        }
        unbrokenUpwardsInteritanceSchemaOccurrenceMatrix.RowNames = rowNames;
        unbrokenUpwardsInteritanceSchemaOccurrenceMatrix.ColumnNames = new[] { ((Permutation)bestIndividual.Data).ToString() };
        Results.Add(new Result(UnbrokenUpwardsInteritanceWildcardOccurrenceMatrixResultName, unbrokenUpwardsInteritanceSchemaOccurrenceMatrix));
      }
      return base.Apply();
    }

    private string TrackUnbrokenAncestorSchemaPath(IntArray schema, GenealogyGraphNode bestIndividual, IList<GenealogyGraphNode>[] ancestors) {
      StringBuilder sb = new StringBuilder();
      for (int i = (int)bestIndividual.Rank.Max(); i >= 0; i--) {
        // check if the schema matches a part of any ancestor in this generation
        var matches = ancestors[(int)i].Where(x => AlgorithmBehaviorHelpers.IsMatch(schema, ((Permutation)x.Data).ToArray()));
        if (matches.Any()) {
          if (i != (int)bestIndividual.Rank.Max()) sb.Append(", ");
          if (matches.Any(x => x.Rank.Contains(i - 0.5))) // mutation occurred
            sb.Append(i + "(MX) <- " + (i - 0.5) + " (CX)");
          else
            sb.Append(i + "(CX)");
        } else break;
      }
      return sb.ToString();
    }
  }
}
