﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes the quality of schemas.
  /// </summary>
  [Item("SchemaQualityAnalyzer", "An operator that analyzes broken inheritance of building blocks.")]
  [StorableClass]
  public sealed class SchemaQualityAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemataParameterName = "Schemata";
    private const string SchemaQualityMatrixParameterName = "SchemaQualityMatrix";
    private const string QualitiesParameterName = "Qualities";
    private const string DistanceMatrixParameterName = "DistanceMatrix";
    private const string GenerationsParameterName = "Generations";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DistanceMatrix>)Parameters[DistanceMatrixParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.ActualValue; }
    }
    public HeuristicLab.Analysis.DataTable Qualities {
      get { return ((HeuristicLab.Analysis.DataTable)ResultsParameter.ActualValue[QualitiesParameterName].Value); }
    }
    #endregion

    [StorableConstructor]
    private SchemaQualityAnalyzer(bool deserializing) : base(deserializing) { }
    private SchemaQualityAnalyzer(SchemaQualityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public SchemaQualityAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<DistanceMatrix>(DistanceMatrixParameterName, "The distance matrix of the TSP."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SchemaQualityAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      var schemata = (ItemArray<IntArray>)Results[SchemataParameterName].Value;

      var schemataMatrix = new Dictionary<IntArray, int[]>(new SchemaEqualityComparer());
      int qualityPoints = 0;
      int worseQualityPoints = 0;
      foreach (var schema in schemata) {
        if (!schemataMatrix.ContainsKey(schema)) {
          var categories = new int[5];
          categories[0] = ExploreGlobalSchemaOccurrence(graph, schema, out qualityPoints, out worseQualityPoints);
          categories[1] = qualityPoints;
          categories[2] = worseQualityPoints;
          categories[3] = (qualityPoints != 0 && worseQualityPoints != 0) ? (int)((qualityPoints / ((double)worseQualityPoints)) * 100.0) : 0;
          categories[4] = ExploreSchemaGradientOverGenerations(graph, schema);
          schemataMatrix.Add(schema, categories);
        }
      }

      // order schemas according to their global occurrence count
      var sortedSchemataMatrix = schemataMatrix.OrderByDescending(x => x.Value[0]).ToDictionary(x => x.Key, x => x.Value);

      var stringMatrix = new StringMatrix(sortedSchemataMatrix.Keys.Count, 6);
      for (int i = 0; i < sortedSchemataMatrix.Keys.Count; i++) {
        var element = sortedSchemataMatrix.ElementAt(i);
        stringMatrix[i, 0] = new Permutation(PermutationTypes.RelativeUndirected, element.Key).ToString();
        for (int j = 0; j < element.Value.Count(); j++) {
          stringMatrix[i, j + 1] = element.Value[j].ToString();
        }
      }
      stringMatrix.ColumnNames = new[] { "Schema", "GlobalSchemaOccurrence", "Better than Avg. Schemata", "Worse than Avg. Schemata", "Ratio", "Gradient" };

      Results.Add(new Result(SchemaQualityMatrixParameterName, stringMatrix));

      return base.Apply();
    }

    private int ExploreSchemaGradientOverGenerations(GenealogyGraph<Permutation> graph, IntArray schema) {
      int[] occurences = new int[GenerationsParameter.ActualValue.Value + 1];
      double gradient = 0.0;

      foreach (var individual in graph.Values) {
        if (AlgorithmBehaviorHelpers.IsSubtour(schema, (Permutation)individual.Data)) {
          foreach (var rank in individual.Rank) {
            occurences[(int)rank]++;
          }
        }
      }

      for (int i = 0; i < occurences.Length - 1; i++) {
        int occurence = occurences[i + 1] - occurences[i];

        if (occurences[i + 1] != 0 && occurences[i] != 0) {
          if (occurence != 0) {
            gradient += occurences[i + 1] - occurences[i];
          } else {
            gradient += 0.5; //TODO: scale this
          }
        }
      }

      return (int)Math.Round(gradient);
    }

    private int ExploreGlobalSchemaOccurrence(GenealogyGraph<Permutation> graph, IntArray schema, out int qualityPoint, out int worstQualityPoint) {
      int occurrences = 0;
      qualityPoint = 0;
      worstQualityPoint = 0;
      foreach (var individual in graph.Values) {
        if (AlgorithmBehaviorHelpers.IsSubtour(schema, (Permutation)individual.Data)) {
          occurrences++;

          Permutation p = (Permutation)individual.Data;
          double pQuality = TSPDistanceMatrixEvaluator.Apply(DistanceMatrix, p);

          //TODO: fix that
          /*if (Qualities.Rows["CurrentAverageQuality"].Values[(int)individual.Rank] > pQuality) {
            qualityPoint += 1;
          } else {
            worstQualityPoint += 1;
          } */
        }
      }
      return occurrences;
    }
  }
}
